#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <xf86drm.h>
#include "nouveau_drm.h"
#include "nouveau_device.h"
#include "nouveau_notifier.h"
#include "nouveau_channel.h"
#include "nouveau_pushbuf.h"
#include "nouveau_bo.h"
#include "nouveau_class.h"

#define NvDmaFB 0xD0FB0001
#define NvDmaTT 0xD0AA0001
#define Nv2D 0xD02D0001
#define NvNotify 0xD0000001

struct nouveau_device *dev = NULL;
struct nouveau_channel *chan = NULL;
struct nouveau_notifier *notifier = NULL;
struct nouveau_grobj *eng2d = NULL;
struct nouveau_bo *bo = NULL;

#define NV84_MODE 1
#define NV90_MODE 0

#define WIDTH 200
#define HEIGHT 200
#define TILING_PARAM 0x00

#if TILING_PARAM == 0x00
#define TILE_WIDTH 32
#define TILE_HEIGHT 4
#elif TILING_PARAM == 0x10
#define TILE_WIDTH 64
#define TILE_HEIGHT 8
#elif TILING_PARAM == 0x20
#define TILE_WIDTH 64
#define TILE_HEIGHT 16
#elif TILING_PARAM == 0x30
#define TILE_WIDTH 64
#define TILE_HEIGHT 32
#elif TILING_PARAM == 0x40
#define TILE_WIDTH 64
#define TILE_HEIGHT 64
#endif

#define PITCH_WIDTH ((WIDTH + 63) & ~63)
#define PITCH_HEIGHT ((HEIGHT + TILE_HEIGHT - 1) & ~(TILE_HEIGHT - 1))

void
get_dst()
{
	int ret = 0, i;
	uint8_t *map;

	if (ret = nouveau_bo_new(dev, NOUVEAU_BO_VRAM | NOUVEAU_BO_TILED, 256, PITCH_WIDTH*PITCH_HEIGHT + 768, &bo)) {
		fprintf(stderr, "nouveau_bo_new failed with %d\n", ret);
		return;
	}

	/* Set destination. */
	BEGIN_RING (chan, eng2d, NV50_2D_DST_FORMAT, 5);
	OUT_RING (chan, NV50_2D_SRC_FORMAT_8BPP);
	OUT_RING (chan, 0);
	OUT_RING (chan, TILING_PARAM);
	OUT_RING  (chan, 1);
	OUT_RING  (chan, 0);

	BEGIN_RING (chan, eng2d, NV50_2D_DST_WIDTH, 4);
	OUT_RING (chan, WIDTH);
	OUT_RING (chan, HEIGHT);
	OUT_RELOCh(chan, bo, 0, NOUVEAU_BO_VRAM | NOUVEAU_BO_WR);
	OUT_RELOCl(chan, bo, 0, NOUVEAU_BO_VRAM | NOUVEAU_BO_WR);

	BEGIN_RING(chan, eng2d, NV50_2D_CLIP_X, 4);
	OUT_RING  (chan, 0);
	OUT_RING  (chan, 0);
	OUT_RING  (chan, WIDTH);
	OUT_RING  (chan, HEIGHT);

	BEGIN_RING(chan, eng2d, NV50_2D_OPERATION, 1);
	OUT_RING  (chan, NV50_2D_OPERATION_SRCCOPY);

	FIRE_RING  (chan);

	/* clear the pixmap */
	nouveau_bo_map(bo, NOUVEAU_BO_RDWR);
	map = bo->map;
	if (!map) {
		fprintf(stderr, "map failed\n");
		return;
	}

	for (i = 0; i < bo->size; i++)
		map[i] = 0x00;

	nouveau_bo_unmap(bo);
}

void
set_pixel(int x, int y, uint32_t pixel)
{
	BEGIN_RING (chan, eng2d, NV50_2D_DST_ADDRESS_HIGH, 2);
	OUT_RELOCh(chan, bo, 0, NOUVEAU_BO_VRAM | NOUVEAU_BO_WR);
	OUT_RELOCl(chan, bo, 0, NOUVEAU_BO_VRAM | NOUVEAU_BO_WR);

	BEGIN_RING(chan, eng2d, 0x580, 3);
	OUT_RING  (chan, 4);
	OUT_RING  (chan, NV50_2D_SRC_FORMAT_8BPP);
	OUT_RING  (chan, pixel);

	BEGIN_RING(chan, eng2d, NV50_2D_RECT_X1, 4);
	OUT_RING  (chan, x);
	OUT_RING  (chan, y);
	OUT_RING  (chan, x + 1);
	OUT_RING  (chan, y + 1);

	FIRE_RING (chan);
}

void
find_pixel (int *offset, uint32_t pixel)
{
	uint8_t *map = NULL;
	int smart_offset = *offset - 2048;
	int smart_end = smart_offset + 4096;

	if (smart_offset < 0)
		smart_offset = 0;
	if (smart_end > bo->size)
		smart_end = bo->size;

	nouveau_bo_map(bo, NOUVEAU_BO_RDWR);
	map = bo->map;
	if (!map) {
		fprintf(stderr, "map failed\n");
		exit(0);
	};

	for ((*offset) = smart_offset; (*offset) < smart_end; (*offset)++)
		if ((map[(*offset)]) == ((pixel) & 0xFF))
			goto out;


	for ((*offset) = 0; (*offset) < bo->size; (*offset)++)
		if ((map[(*offset)]) == ((pixel) & 0xFF))
			goto out;

out:
	map[(*offset)] = 0x00;
	nouveau_bo_unmap(bo);
}


static int pattern_nv9x[] = { 0, 1, 2, 3, /**/ 0, 1, 2, 3, /**/ 1, 2, 3, /**/ 0, 2, 3, /**/ 
					0, 1, 3, /**/ 0, 1, 2, /**/ 1, 2, 3, /**/ 0, 2, 3, /**/ 0, 1, 3, /**/ 0, 1, 2 /**/};
static int pattern_nv84[] = { 4, /**/ 5, 5, /**/ 4, 5, /**/ 4, 4, /**/ 5, 5, /**/ 4, 4, /**/ 5, 4, /**/ 5, 5, /**/ 4 };
static int tile_offset[] = {
					0, 0, 0, 0, 0, 0, 0, 0, /* type 0 */
					2, 2, 2, 2, 2, 2, -6, -6, /* type 1 */
					4, 4, 4, 4, -4, -4, -4, -4, /* type 2 */
					6, 6, -2, -2, -2, -2, -2, -2, /* type 3 */
					0, 0, 0, 0, 2, 2, -2, -2, /* type 4 */
					2, 2, -2, -2, 0, 0, 0, 0, /* type 5 */
					};

int
recurse_offset(int offset, int tile_height, int x, int y, int old_tile_width, int first)
{
	int tile_width = 0;
	if (tile_height == 2)
		tile_width = 8;
	else if (tile_height == 4)
		tile_width = 32;
	else
		tile_width = 64;

	offset += ((x/tile_width) + (y/tile_height) * (old_tile_width/tile_width)) * (tile_height *  tile_width);

	if (tile_height > 2)
		offset = recurse_offset(offset, tile_height/2, x % tile_width, y % tile_height, tile_width, 0);
	else
		offset += (y % tile_height) * tile_width + (x % tile_width);

	if (first && NV84_MODE) {
		unsigned int suboffset, suboffset2, suboffset3;

		/* 128 byte blocks within a larger block of 16384 bytes */
		suboffset = (offset & 0x3FFF) & ~0x7F;
		/* now we have 128 blocks */
		suboffset >>= 7;

		/* now we have 16 rows */
		suboffset2 = suboffset >> 3;
		/* tile within row */
		suboffset3 = suboffset & 0x7;

		offset += tile_offset[suboffset3 + 8*pattern_nv84[suboffset2]] * 128;
	}

	if (first && NV90_MODE) {
		unsigned int suboffset, suboffset2, suboffset3;

		/* 128 byte blocks within a larger block of 32768 bytes */
		/* the first block misses the first "0, 1, 2, 3" sequence, so we add an extra 4096 bytes offset. */
		suboffset = ((offset + 4096) & 0x7FFF) & ~0x7F;
		/* now we have 256 blocks */
		suboffset >>= 7;

		/* now we have 32 rows */
		suboffset2 = suboffset >> 3;
		/* tile within row */
		suboffset3 = suboffset & 0x7;

		offset += tile_offset[suboffset3 + 8*pattern_nv9x[suboffset2]] * 128;
	}

	return offset;
}

int
main(int argc, char **argv)
{
	int ret = 0;
	int x = 0, y = 0;
	int offset = 0, offset_expected = 0;

	if (ret = nouveau_device_open(&dev, NULL)) {
		fprintf(stderr, "nouveau_device_open failed\n");
		return ret;
	}

	if (ret = nouveau_channel_alloc(dev, NvDmaFB, NvDmaTT, &chan)) {
		fprintf(stderr, "nouveau_channel_alloc failed\n");
		return ret;
	}

	if (ret = nouveau_notifier_alloc(chan, NvNotify, 1, &notifier)) {
		fprintf(stderr, "nouveau_notifier_alloc failed\n");
		return ret;
	}

	/* 2D object */
	if (ret = nouveau_grobj_alloc(chan, Nv2D, 0x502d, &eng2d)) {
		fprintf(stderr, "nouveau_grobj_alloc failed\n");
		return ret;
	}

	BEGIN_RING (chan, eng2d, 0x180, 3);
	OUT_RING  (chan, notifier->handle);
	OUT_RING  (chan, chan->vram->handle);
	OUT_RING  (chan, chan->vram->handle);

	/* Magics from nv, no clue what they do, but at least some
	 * of them are needed to avoid crashes.
	 */
	BEGIN_RING(chan, eng2d, 0x260, 1);
	OUT_RING  (chan, 1);
	BEGIN_RING(chan, eng2d, 0x290, 1);
	OUT_RING  (chan, 1);
	BEGIN_RING(chan, eng2d, 0x29c, 1);
	OUT_RING  (chan, 0);
	BEGIN_RING(chan, eng2d, 0x58c, 1);
	OUT_RING  (chan, 0x111);

	/* Now we are ready for the real stuff. */
	get_dst();
	if (!bo) {
		fprintf(stderr, "get_dst failed\n");
		return -ENOMEM;
	}

	for (y = 0; y < HEIGHT; y++) {
		for (x = 0; x < WIDTH; x++) {
			set_pixel(x, y, 0xFFFFFFFF);
			offset_expected = recurse_offset(0, TILE_HEIGHT, x, y, PITCH_WIDTH, 1);
			offset = offset_expected;
			find_pixel(&offset, 0xFFFFFFFF);
			if (offset_expected != offset) {
				printf("offset %d offset_expected %d\n", offset, offset_expected);
				printf("x %d y %d\n", x, y);
			}
		}
	}

	nouveau_device_close(&dev);

	return ret;
}
